// bebot
// Copyright (C) 2023 Brian Tarricone <brian@tarricone.org>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use core::fmt;

use crate::config::PublishEvent;

pub trait GitlabEventExt {
    fn should_publish(&self, publish_events: &[PublishEvent]) -> bool;
    fn project(&self) -> &Project;
    fn r#ref(&self) -> Option<&str>;
    fn user(&self) -> &str;
    fn titles(&self) -> Vec<String>;
}

#[derive(Debug, Deserialize)]
pub struct User {
    name: String,
}

#[derive(Debug, Deserialize)]
pub struct Project {
    pub name: String,
    pub web_url: String,
    pub namespace: String,
    pub path_with_namespace: String,
    pub default_branch: String,
}

#[derive(Debug, Deserialize)]
pub struct Repository {
    pub name: String,
    pub url: String,
}

#[derive(Debug, Deserialize)]
pub struct Commit {
    pub id: String,
    pub title: String,
    pub url: String,
}

#[derive(PartialEq, Debug, Deserialize)]
#[serde(rename_all = "snake_case")]
pub enum MergeRequestAction {
    Open,
    Close,
    Reopen,
    Update,
    Approve,
    Unapprove,
    Merged,
    Approval,
    Unapproval,
    #[serde(other)]
    Other,
}

impl fmt::Display for MergeRequestAction {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "{}",
            match self {
                MergeRequestAction::Open => "opened",
                MergeRequestAction::Close => "closed",
                MergeRequestAction::Reopen => "reopened",
                MergeRequestAction::Update => "updated",
                MergeRequestAction::Approve => "approved",
                MergeRequestAction::Unapprove => "unapproved",
                MergeRequestAction::Approval => "approval",
                MergeRequestAction::Unapproval => "unapproval",
                MergeRequestAction::Merged => "merged",
                MergeRequestAction::Other => "other",
            }
        )
    }
}

#[derive(Debug, Deserialize)]
pub struct MergeRequestObjectAttributes {
    pub iid: u64,
    pub target_branch: String,
    pub source_branch: String,
    pub title: String,
    pub url: String,
    pub source: Project,
    pub target: Project,
    pub action: MergeRequestAction,
}

#[derive(PartialEq, Debug, Deserialize)]
#[serde(rename_all = "snake_case")]
pub enum PipelineStatus {
    Created,
    WaitingForResource,
    Preparing,
    Pending,
    Running,
    Success,
    Failed,
    Canceled,
    Skipped,
    Manual,
    Scheduled,
    #[serde(other)]
    Other,
}

impl fmt::Display for PipelineStatus {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "{}",
            match self {
                PipelineStatus::Created => "created",
                PipelineStatus::WaitingForResource => "waiting for resource",
                PipelineStatus::Preparing => "preparing",
                PipelineStatus::Pending => "pending",
                PipelineStatus::Running => "running",
                PipelineStatus::Success => "succeeded",
                PipelineStatus::Failed => "failed",
                PipelineStatus::Canceled => "canceled",
                PipelineStatus::Skipped => "skipped",
                PipelineStatus::Manual => "manual",
                PipelineStatus::Scheduled => "scheduled",
                PipelineStatus::Other => "other",
            }
        )
    }
}

#[derive(Debug, Deserialize)]
pub struct PipelineObjectAttributes {
    pub name: Option<String>,
    pub r#ref: String,
    pub status: PipelineStatus,
    pub url: String,
}

#[derive(Debug, Deserialize)]
pub struct PipelineMergeRequest {
    pub title: String,
}

#[derive(PartialEq, Debug, Deserialize)]
#[serde(rename_all = "snake_case")]
pub enum IssueAction {
    Open,
    Close,
    Reopen,
    Update,
    #[serde(other)]
    Other,
}

impl fmt::Display for IssueAction {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "{}",
            match self {
                IssueAction::Open => "opened",
                IssueAction::Close => "closed",
                IssueAction::Reopen => "reopened",
                IssueAction::Update => "updated",
                IssueAction::Other => "other",
            }
        )
    }
}

#[derive(Debug, Deserialize)]
pub struct IssueObjectAttributes {
    pub id: u32,
    pub title: String,
    pub action: IssueAction,
    pub url: String,
}

#[derive(Debug, Deserialize)]
#[serde(tag = "object_kind")]
pub enum GitlabEvent {
    #[serde(rename = "push")]
    Push {
        event_name: String,
        before: String,
        after: String,
        r#ref: String,
        ref_protected: bool,
        checkout_sha: String,
        user_id: u64,
        user_name: String,
        user_username: String,
        user_email: String,
        user_avatar: Option<String>,
        project_id: u64,
        project: Project,
        repository: Repository,
        commits: Vec<Commit>,
        total_commits_count: u64,
    },
    #[serde(rename = "tag_push")]
    TagPush {
        event_name: String,
        before: String,
        after: String,
        r#ref: String,
        ref_protected: bool,
        checkout_sha: String,
        user_id: u64,
        user_name: String,
        user_avatar: Option<String>,
        project_id: u64,
        project: Project,
        repository: Repository,
        commits: Vec<Commit>,
        total_commits_count: u64,
    },
    #[serde(rename = "issue")]
    Issue {
        user: User,
        project: Project,
        object_attributes: IssueObjectAttributes,
    },
    #[serde(rename = "merge_request")]
    MergeRequest {
        user: User,
        project: Project,
        repository: Repository,
        object_attributes: MergeRequestObjectAttributes,
    },
    #[serde(rename = "pipeline")]
    Pipeline {
        object_attributes: PipelineObjectAttributes,
        merge_request: Option<PipelineMergeRequest>,
        user: User,
        project: Project,
    },
    #[serde(other)]
    Other,
}

macro_rules! find_publish_event {
    ($cfgs:expr, $case:pat) => {
        $cfgs.iter().find(|cfg| match cfg {
            $case => true,
            _ => false,
        })
    };
}

impl GitlabEventExt for GitlabEvent {
    fn should_publish(&self, publish_events: &[PublishEvent]) -> bool {
        match self {
            GitlabEvent::Push { r#ref, .. } => {
                if let Some(PublishEvent::Push { branches }) =
                    find_publish_event!(publish_events, PublishEvent::Push { .. })
                {
                    match branches {
                        None => true,
                        Some(branches) => {
                            let refname = parse_ref(r#ref);
                            branches.iter().any(|branch| branch.find(&refname).is_some())
                        }
                    }
                } else {
                    false
                }
            }
            GitlabEvent::TagPush { .. } => find_publish_event!(publish_events, PublishEvent::TagPush { .. }).is_some(),
            GitlabEvent::Issue { object_attributes, .. } => {
                if let Some(PublishEvent::Issues { actions }) =
                    find_publish_event!(publish_events, PublishEvent::Issues { .. })
                {
                    match actions {
                        None => true,
                        Some(actions) => actions.contains(&object_attributes.action),
                    }
                } else {
                    false
                }
            }
            GitlabEvent::MergeRequest { object_attributes, .. } => {
                if let Some(PublishEvent::MergeRequest { actions }) =
                    find_publish_event!(publish_events, PublishEvent::MergeRequest { .. })
                {
                    match actions {
                        None => true,
                        Some(actions) => actions.contains(&object_attributes.action),
                    }
                } else {
                    false
                }
            }
            GitlabEvent::Pipeline { object_attributes, .. } => {
                if let Some(PublishEvent::Pipeline { statuses }) =
                    find_publish_event!(publish_events, PublishEvent::Pipeline { .. })
                {
                    match statuses {
                        None => true,
                        Some(statuses) => statuses.contains(&object_attributes.status),
                    }
                } else {
                    false
                }
            }
            GitlabEvent::Other => false,
        }
    }

    fn project(&self) -> &Project {
        match self {
            GitlabEvent::Push { project, .. } => project,
            GitlabEvent::TagPush { project, .. } => project,
            GitlabEvent::Issue { project, .. } => project,
            GitlabEvent::MergeRequest { project, .. } => project,
            GitlabEvent::Pipeline { project, .. } => project,
            GitlabEvent::Other => unreachable!("Unsupported event type"),
        }
    }

    fn r#ref(&self) -> Option<&str> {
        match self {
            GitlabEvent::Push { r#ref, .. } => Some(r#ref),
            GitlabEvent::TagPush { r#ref, .. } => Some(r#ref),
            GitlabEvent::Issue { .. } => None,
            GitlabEvent::MergeRequest { object_attributes, .. } => Some(&object_attributes.target_branch),
            GitlabEvent::Pipeline { object_attributes, .. } => Some(&object_attributes.r#ref),
            GitlabEvent::Other => unreachable!("Unsupported event type"),
        }
    }

    fn user(&self) -> &str {
        match self {
            GitlabEvent::Push { user_name, .. } => user_name,
            GitlabEvent::TagPush { user_name, .. } => user_name,
            GitlabEvent::Issue { user, .. } => &user.name,
            GitlabEvent::MergeRequest { user, .. } => &user.name,
            GitlabEvent::Pipeline { user, .. } => &user.name,
            GitlabEvent::Other => unreachable!("Unsupported event type"),
        }
    }

    fn titles(&self) -> Vec<String> {
        fn find_commit<'a>(commits: &'a [Commit], sha: &str) -> Option<&'a Commit> {
            commits.iter().find(|commit| commit.id == sha)
        }

        match self {
            GitlabEvent::Push {
                after,
                project,
                commits,
                ..
            } => {
                const MAX_COMMITS: usize = 15; // TODO: make configurable
                commits.iter().fold(Vec::new(), |mut titles, commit| {
                    if titles.len() < MAX_COMMITS {
                        titles.push(format!("[**pushed** {}]({})", commit.title, commit.url));

                        if titles.len() == MAX_COMMITS {
                            titles.push(format!(
                                "[**pushed** {} more commits]({}/-/compare/{}...{})",
                                commits.len() - MAX_COMMITS,
                                project.web_url,
                                commit.id,
                                after
                            ));
                        }
                    }

                    titles
                })
            }
            GitlabEvent::TagPush {
                r#ref,
                checkout_sha,
                project,
                commits,
                ..
            } => {
                let title = format!(
                    "**tagged** {}",
                    find_commit(commits, checkout_sha)
                        .map(|commit| &commit.title)
                        .unwrap_or(checkout_sha)
                );
                let url = format!("{}/-/tags/{}", project.web_url, parse_ref(r#ref));
                vec![markdown_link(&title, &url)]
            }
            GitlabEvent::Issue { object_attributes, .. } => {
                if object_attributes.action != IssueAction::Other {
                    let title = format!(
                        "Issue #{} **{}**: {}",
                        object_attributes.id, object_attributes.action, object_attributes.title
                    );
                    vec![markdown_link(&title, &object_attributes.url)]
                } else {
                    vec![]
                }
            }
            GitlabEvent::MergeRequest { object_attributes, .. } => {
                if object_attributes.action != MergeRequestAction::Other {
                    let title = format!(
                        "MR !{} **{}**: {}",
                        object_attributes.iid, object_attributes.action, object_attributes.title
                    );
                    vec![markdown_link(&title, &object_attributes.url)]
                } else {
                    vec![]
                }
            }
            GitlabEvent::Pipeline {
                object_attributes,
                merge_request,
                ..
            } => {
                if object_attributes.status != PipelineStatus::Other {
                    let title = object_attributes
                        .name
                        .as_ref()
                        .cloned()
                        .or(merge_request.as_ref().map(|mr| mr.title.clone()))
                        .iter()
                        .fold(format!("Pipeline **{}**", object_attributes.status), |accum, title| {
                            format!("{}: {}", accum, title)
                        });
                    vec![markdown_link(&title, &object_attributes.url)]
                } else {
                    vec![]
                }
            }
            GitlabEvent::Other => unreachable!("Unsupported event type"),
        }
    }
}

#[inline]
fn markdown_link(title: &String, url: &String) -> String {
    format!("[{}]({})", title, url)
}

pub fn parse_ref(r#ref: &str) -> String {
    if r#ref.starts_with("refs/") {
        let parts = r#ref.split('/').collect::<Vec<_>>();
        if parts.len() > 2 {
            parts.into_iter().skip(2).collect::<Vec<_>>().join("/").to_string()
        } else {
            r#ref.to_string()
        }
    } else {
        r#ref.to_string()
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use regex::Regex;
    use std::{fs::File, io::BufReader};

    fn load_test_data(name: &str) -> anyhow::Result<GitlabEvent> {
        let f = File::open(format!("{}/test-data/{}.json", env!("CARGO_MANIFEST_DIR"), name))?;
        let r = BufReader::new(f);
        let event: GitlabEvent = serde_json::from_reader(r)?;
        Ok(event)
    }

    #[test]
    pub fn parse_push_event() -> anyhow::Result<()> {
        let event = load_test_data("push-event")?;

        match &event {
            GitlabEvent::Push {
                event_name,
                before,
                after,
                r#ref,
                checkout_sha,
                user_username,
                project,
                repository,
                total_commits_count,
                ..
            } => {
                assert_eq!(event_name, "push");
                assert_eq!(before, "95790bf891e76fee5e1747ab589903a6a1f80f22");
                assert_eq!(after, "da1560886d4f094c3e6c9ef40349f7d38b5d27d7");
                assert_eq!(r#ref, "refs/heads/master");
                assert_eq!(checkout_sha, "da1560886d4f094c3e6c9ef40349f7d38b5d27d7");
                assert_eq!(user_username, "jsmith");
                assert_eq!(project.name, "Diaspora");
                assert_eq!(project.namespace, "Mike");
                assert_eq!(repository.name, "Diaspora");
                assert_eq!(repository.url, "git@example.com:mike/diaspora.git");
                assert_eq!(*total_commits_count, 4);
            }
            _ => panic!("not a push event"),
        };

        let publish_events = vec![PublishEvent::Push { branches: None }];
        assert!(event.should_publish(&publish_events));

        let publish_events = vec![PublishEvent::Push {
            branches: Some(vec![Regex::new(r"^master$").unwrap()]),
        }];
        assert!(event.should_publish(&publish_events));

        let publish_events = vec![PublishEvent::Push {
            branches: Some(vec![Regex::new(r"^mas").unwrap()]),
        }];
        assert!(event.should_publish(&publish_events));

        let publish_events = vec![PublishEvent::Push {
            branches: Some(vec![Regex::new(r"^foobar$").unwrap()]),
        }];
        assert!(!event.should_publish(&publish_events));

        let publish_events = vec![PublishEvent::Pipeline { statuses: None }];
        assert!(!event.should_publish(&publish_events));

        Ok(())
    }

    #[test]
    pub fn parse_tag_push_event() -> anyhow::Result<()> {
        let event = load_test_data("tag-push-event")?;

        match &event {
            GitlabEvent::TagPush {
                event_name,
                before,
                after,
                r#ref,
                checkout_sha,
                user_name,
                project,
                repository,
                total_commits_count,
                ..
            } => {
                assert_eq!(event_name, "tag_push");
                assert_eq!(before, "0000000000000000000000000000000000000000");
                assert_eq!(after, "82b3d5ae55f7080f1e6022629cdb57bfae7cccc7");
                assert_eq!(r#ref, "refs/tags/v1.0.0");
                assert_eq!(checkout_sha, "82b3d5ae55f7080f1e6022629cdb57bfae7cccc7");
                assert_eq!(user_name, "John Smith");
                assert_eq!(project.name, "Example");
                assert_eq!(project.namespace, "Jsmith");
                assert_eq!(repository.name, "Example");
                assert_eq!(*total_commits_count, 0);
            }
            _ => panic!("not a tag push event"),
        };

        let publish_events = vec![PublishEvent::TagPush];
        assert!(event.should_publish(&publish_events));

        let publish_events = vec![PublishEvent::Push { branches: None }];
        assert!(!event.should_publish(&publish_events));

        Ok(())
    }

    #[test]
    pub fn parse_issue_event() -> anyhow::Result<()> {
        let event = load_test_data("issue-event")?;

        match &event {
            GitlabEvent::Issue {
                user,
                object_attributes,
                ..
            } => {
                assert_eq!(user.name, "Administrator");
                assert_eq!(object_attributes.id, 301);
                assert_eq!(object_attributes.action, IssueAction::Open);
            }
            _ => panic!("not an issue event"),
        }

        let publish_events = vec![PublishEvent::Issues { actions: None }];
        assert!(event.should_publish(&publish_events));

        let publish_events = vec![PublishEvent::Issues {
            actions: Some(vec![IssueAction::Open]),
        }];
        assert!(event.should_publish(&publish_events));

        let publish_events = vec![PublishEvent::Issues {
            actions: Some(vec![IssueAction::Close]),
        }];
        assert!(!event.should_publish(&publish_events));

        let publish_events = vec![PublishEvent::Push { branches: None }];
        assert!(!event.should_publish(&publish_events));

        Ok(())
    }

    #[test]
    pub fn parse_merge_request_event() -> anyhow::Result<()> {
        let event = load_test_data("merge-request-event")?;

        match &event {
            GitlabEvent::MergeRequest {
                user,
                object_attributes,
                ..
            } => {
                assert_eq!(user.name, "Administrator");
                assert_eq!(object_attributes.action, MergeRequestAction::Open);
                assert_eq!(object_attributes.title, "MS-Viewport");
            }
            _ => panic!("not a merge request event"),
        };

        let publish_events = vec![PublishEvent::MergeRequest { actions: None }];
        assert!(event.should_publish(&publish_events));

        let publish_events = vec![PublishEvent::MergeRequest {
            actions: Some(vec![MergeRequestAction::Open]),
        }];
        assert!(event.should_publish(&publish_events));

        let publish_events = vec![PublishEvent::MergeRequest {
            actions: Some(vec![MergeRequestAction::Close]),
        }];
        assert!(!event.should_publish(&publish_events));

        let publish_events = vec![PublishEvent::Push { branches: None }];
        assert!(!event.should_publish(&publish_events));

        Ok(())
    }

    #[test]
    pub fn parse_pipeline_event() -> anyhow::Result<()> {
        let event = load_test_data("pipeline-event")?;

        match &event {
            GitlabEvent::Pipeline {
                object_attributes,
                merge_request,
                user,
                ..
            } => {
                assert_eq!(object_attributes.name, Some("Pipeline for branch: master".to_string()));
                assert_eq!(object_attributes.r#ref, "master");
                assert_eq!(object_attributes.status, PipelineStatus::Failed);
                assert_eq!(merge_request.as_ref().unwrap().title, "Test");
                assert_eq!(user.name, "Administrator");
            }
            _ => panic!("not a pipeline event"),
        };

        let publish_events = vec![PublishEvent::Pipeline { statuses: None }];
        assert!(event.should_publish(&publish_events));

        let publish_events = vec![PublishEvent::Pipeline {
            statuses: Some(vec![PipelineStatus::Failed]),
        }];
        assert!(event.should_publish(&publish_events));

        let publish_events = vec![PublishEvent::Pipeline {
            statuses: Some(vec![PipelineStatus::Success]),
        }];
        assert!(!event.should_publish(&publish_events));

        let publish_events = vec![PublishEvent::Push { branches: None }];
        assert!(!event.should_publish(&publish_events));

        Ok(())
    }
}
